#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# afb 2/96, etter 6/96, afb 12/1999, 10/2002
#------------------------------------------------------------------------------
# Implementation of all accessing operations on base of a cache
#------------------------------------------------------------------------------

package TBD::CachingTable;
use Carp;
use English;
use TBI;
use strict;
use warnings;

our $VERSION = "0.09";
our @ISA = qw(TBI);

=head1 NAME

TBD::CachingTable -- implementation of all accessing operations on base of a cache

=head1 SYNOPSIS

   use TBD::CachingTable;
   our @ISA = qw(... TBD::CachingTable ...);

=head1 DESCRIPTION

TBD::CachingTable implements the accessing operations on base of
a cache that must be initialized and saved by other modules.
Following internal variables are expected to be properly
initialized:

=over 4

=item $table->{table}

The cache itself which is represented as hash whose values are references
to hash tables which represent individual records of the table.

=item $table->{loaded}

Any accesses are denied if this variable is not set to true.

=item $table->{locked}

Any modifying accesses are denied if this variable is not set to true.

=item $table->{ofs}

In case of keys consisting of multiple fields this separator is
used to construct keys of $table->{table}.

=item $table->{ifs}

Is used to split joined keys consisting of multiple fields.

=back

Following operations are implemented:
fetch, exists, get, keys, select, add, delete, and modify.

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/borchert/">Andreas Borchert</A>
with contributions of
<A HREF="http://www.mathematik.uni-ulm.de/sai/etter/">Mathias Etter</A>.

=for text
Andreas Borchert with contributions of Mathias Etter.

=for man
Andreas Borchert with contributions of Mathias Etter.

=cut

sub _mkkey {
   my ($self, $key) = @_;
   my @keyfields = $self->keyfields;
   if (@keyfields > 1) {
      croak "invalid key" if ref($key) ne "HASH";
   }
   if (ref($key) eq "HASH") {
      my %key = %{$key};
      $key = join($self->{ofs}, @key{@keyfields});
   }
   return $key;
}

sub fetch {
   my ($self, $key) = @_;

   croak "table is not loaded yet" unless $self->{loaded};
   $key = $self->_mkkey($key);
   croak "invalid key: $key" unless exists $self->{table}->{$key};
   return %{$self->{table}->{$key}};
}

sub exists {
   my ($self, $key) = @_;

   croak "table is not loaded yet" unless $self->{loaded};
   $key = $self->_mkkey($key);
   return exists $self->{table}->{$key};
}

sub get {
   my ($self, $key, @fieldnames) = @_;
   my (@fields);

   croak "table is not loaded yet" unless $self->{loaded};
   $key = $self->_mkkey($key);
   foreach my $field (@fieldnames) {
      push(@fields, $self->{table}->{$key}->{$field});
   }
   return @fields;
}

sub getfield {
   my ($self, $key, $fieldname) = @_;
   croak "table is not loaded yet" unless $self->{loaded};
   $key = $self->_mkkey($key);
   return $self->{table}->{$key}->{$fieldname};
}

sub _genkeys {
   my ($self, @keys) = @_;

   my @keyfields = $self->keyfields;
   return @keys if @keyfields == 1;
   return map {
      do {
	 my %key = ();
	 my @fields = @keyfields;
	 foreach my $keyfield (split /$self->{ifs}/) {
	    $key{shift @fields} = $keyfield;
	 }
	 \%key;
      }
   } @keys;
}

sub keys {
   my $self = shift;

   croak "table is not loaded yet" unless $self->{loaded};
   return $self->_genkeys(CORE::keys %{$self->{table}});
}

sub select {
   my ($self, %equations) = @_;
   croak "table is not loaded yet" unless $self->{loaded};
   if (%equations) {
      my @keys = ();
      my ($key, $fields);
      LOOP: while (($key, $fields) = each(%{$self->{table}})) {
	 foreach my $fieldname (CORE::keys %equations) {
	    next LOOP if $fields->{$fieldname} ne $equations{$fieldname}
	 }
	 push(@keys, $key);
      }
      return $self->_genkeys(@keys);
   } else {
      # no equations given, so return all keys
      return $self->_genkeys(CORE::keys %{$self->{table}});
   }
}

sub add {
   my ($self, $key, %fields) = @_;
   croak "table is not locked yet" unless $self->{locked};
   $key = $self->_mkkey($key);
   croak "there is already an entry with the given key"
      if exists $self->{table}->{$key};
   $self->{table}->{$key} = \%fields;
}

sub delete {
   my ($self, $key) = @_;

   croak "table is not locked yet" unless $self->{locked};
   $key = $self->_mkkey($key);
   delete $self->{table}->{$key};
}

sub modify {
   my ($self, $key, %fields) = @_;
   my ($field, $value);

   croak "table is not locked yet" unless $self->{locked};
   $key = $self->_mkkey($key);
   croak "unknown key" unless exists $self->{table}->{$key};
   while (($field, $value) = each(%fields)) {
      $self->{table}->{$key}->{$field} = $value;
   }
   return 1;
}

1;
