#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# afb 6/98
#------------------------------------------------------------------------------
# support of tables of relational databases on base of DBI
#------------------------------------------------------------------------------
# $Id$
#------------------------------------------------------------------------------
# $Log$
#------------------------------------------------------------------------------

package TBD::DBI;

use Carp;
use DBI;
use TBI;
use strict;
use warnings;

our $VERSION = "0.03";
our @ISA = qw(TBI);

=head1 NAME

TBD::DBI - support of tables of relational databases (via DBI)

=head1 SYNOPSIS

   $t = new TBD::DBI::MyRelationalDatabase(
      database => $data_source, # 1st parameter of DBI->connect
      login => $login,          # 2nd parameter of DBI->connect
      auth => $auth,            # 3rd parameter of DBI->connect
      table => $tables[0],      # name of the database table
      tablename => $tables[0],  # to be registered under this name
   );

=head1 DESCRIPTION

I<TBD::DBI> supports all I<DBI> implementations for I<TBI> but is by
itself not a full implementation because there is no portable method
to extract meta informations from a relational database. Therefore,
a derived module of I<TBD::DBI> is required that provides following methods:

=over 4

=item initialize1

(or initialize2 -- that does not matter) is expected to set
the component I<keyfields>.

=item tables

is a class method that is expected to return a list of all table names
for a given database handler of I<DBI>.

=back

Hence, implementations are expected to follow this pattern:

   package TBD::DBI::MyRelationalDatabase;

   use DBI;
   use TBD::DBI;
   use strict;
   use warnings;

   our @ISA = qw(TBD::DBI);

   sub tables {
      my ($package, $db) = @_;
      # extract all table names of $db
      return @tables;
   }

   sub initialize1 {
      my ($self) = @_;

      # set $self->{keyfields}

      $self->SUPER::initialize2();
   }

Following components of the table objects may be used by I<initialize1>
or I<initialize2> by a derived module:

=over 4

=item db

database handler of I<DBI>.

=item table

is the name of the database table. While this is in most cases identical
to the table name it could be different.

=back

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/borchert/">Andreas Borchert</A>

=for text
Andreas Borchert

=for man
Andreas Borchert

=cut

# foreach combination of server and database we will maintain
# maximal one connection:

my %connections = ();  # key is a combination of database, login, and auth

sub initialize1 {
   my ($self, %attributes) = @_;

   $self->SUPER::initialize1(%attributes);

   foreach my $field (qw(database login auth table tablename)) {
      croak "Undefined $field" unless defined $attributes{$field};
      $self->{$field} = $attributes{$field};
   }
   my $database = $self->{database};
   my $login = $self->{login};
   my $auth = $self->{auth};
   my $key = "$database/$login/$auth";
   $self->{key} = $key;
   my $table = $self->{table};

   my $db;
   if (defined($connections{$key})) {
      ++ $connections{$key}->{refs};
      $db = $connections{$key}->{handle};
   } else {
      $db = DBI->connect($database, $login, $auth);
      croak "unable to connect to database $database"
	 unless defined $db;
      $connections{$key} = {
	 handle => $db,
	 refs => 1,
      };
      $db->{RaiseError} = 1;
      if ($TBI::AUTO_REG) {
	 foreach my $table (ref($self)->tables($db)) {
	    next if TBI->exists($table);
	    TBI->register($table, ref($self), {
	       database => $database,
	       login => $login,
	       auth => $auth,
	       table => $table,
	       tablename => $table,
	    });
	 }
      }
   }

   my $st = $db->prepare("select * from $table");
   $st->execute();
   $self->{fields} = [@{$st->{NAME}}];
   $st->finish;

   $self->{db} = $db;
}

sub initialize2 {
   my ($self) = @_;
   croak "Key fields not defined" unless ref($self->{keyfields});

   $self->{valid_fieldnames} = {};
   foreach my $field (@{$self->{fields}}) {
      ${$self->{valid_fieldnames}}{$field} = 1;
   }

   # check that all key field names are valid
   foreach my $keyfield (@{$self->{keyfields}}) {
      croak "Unknown field name: $keyfield"
	 unless defined($self->{valid_fieldnames}->{$keyfield});
   }

   my @clauses = ();
   foreach my $keyfield (@{$self->{keyfields}}) {
      push(@clauses, "$keyfield = ?");
   }
   $self->{byKey} = "where " . join(" and ", @clauses);
   $self->{select} = "select * from " . $self->{table} . " ";
   $self->{select_byKey} = $self->{db}->prepare(
      $self->{select} . $self->{byKey});

   my @keyfields = @{$self->{keyfields}};
   my $keyfields = join(", ", @keyfields);
   $self->{select_keys} = $self->{db}->prepare(qq{
      select $keyfields from $self->{table}
   });

   my $tuple = join(", ", ('?') x @{$self->{fields}});
   $self->{insert} = $self->{db}->prepare(qq{
      insert into $self->{table} values ($tuple)
   });

   $self->{delete} = $self->{db}->prepare(qq{
      delete from $self->{table} $self->{byKey}
   });

   $self->SUPER::initialize2();
}

sub close {
   my ($self) = @_;

   return unless defined $self->{db};

   # finish all statement handles created during initialize2 to avoid warnings
   foreach my $handle (qw(select_byKey select_keys insert delete)) {
      $self->{$handle}->finish if defined $self->{$handle};
   }

   undef $self->{db};
   my $key = $self->{key};
   if (--$connections{$key}->{refs} == 0) {
      $connections{$key}->{handle}->disconnect;
      delete $connections{$key};
   }
   $self->SUPER::close();
}

# return list of tables

sub tables {
   my ($self) = @_;

   return $self->{db}->tables();
}

# convert hash of key values into an ordered list of key values
# that may be used as place holders for $self->{byKey}

sub order_keys {
   my ($self, $key) = @_;

   if (ref($key)) {
      my @keys = ();
      foreach my $keyfield (@{$self->{keyfields}}) {
	 croak "missing key field: $keyfield"
	    unless defined $key->{$keyfield};
	 push(@keys, $key->{$keyfield});
      }
      return @keys;
   } else {
      croak "invalid key value" unless @{$self->{keyfields}} == 1;
      return ($key);
   }
}

sub fetch {
   my ($self, $key) = @_;

   my @keys = $self->order_keys($key);
   $self->{select_byKey}->execute(@keys);
   my $record = $self->{select_byKey}->fetch;
   return undef unless defined $record;

   my %fields = (); my $index = 0;
   foreach my $fieldname (@{$self->{fields}}) {
      $fields{$fieldname} = $record->[$index++];
   }
   return %fields;
}

sub exists {
   my ($self, $key) = @_;

   my @keys = $self->order_keys($key);
   my $nofrecords = $self->{select_byKey}->execute(@keys);
   return defined($nofrecords) && $nofrecords > 0;
}

sub keys {
   my $self = shift;

   my $st = $self->{select_keys};
   $st->execute();
   return $self->return_keys($st);
}

sub return_keys {
   my ($self, $st) = @_;

   my @keys = (); my $record;
   if (@{$self->{keyfields}} == 1) {
      while (defined($record = $st->fetch())) {
	 push(@keys, $record->[0]);
      }
   } else {
      while (defined($record = $st->fetchrow_hashref())) {
	 push(@keys, {%{$record}});
      }
   }
   return @keys;
}

sub select {
   my ($self, %clauses) = @_;

   my @clauses = (); my @values = ();
   my ($field, $value);
   while (($field, $value) = each %clauses) {
      push(@clauses, "$field = ?");
      push(@values, $value);
   }
   my @keyfields = @{$self->{keyfields}};
   my $keyfields = join(", ", @keyfields);
   my $st = $self->{db}->prepare(
      "select $keyfields from $self->{table} " .
      " where " . join(" and ", @clauses)
   );
   $st->execute(@values);
   my @keys = $self->return_keys($st);
   $st->finish();
   return @keys;
}

sub hash2tuple {
   my ($self, %hash) = @_;

   my @tuple = ();
   foreach my $fieldname (@{$self->{fields}}) {
      push(@tuple, $hash{$fieldname});
   }
   return @tuple;
}

sub add {
   my ($self, $key, %fields) = @_;
   if (ref($key)) {
      %fields = (%fields, %{$key});
   } else {
      $fields{$self->{keyfields}->[0]} = $key;
   }
   $self->{insert}->execute($self->hash2tuple(%fields));
}

sub delete {
   my ($self, $key) = @_;

   $self->{delete}->execute($self->order_keys($key));
}

sub modify {
   my ($self, $key, %fields) = @_;

   my @updates = (); my @values = ();
   while (my($field, $value) = each(%fields)) {
      push(@updates, "$field = ?");
      push(@values, $value);
   }
   my $updates = join(", ", @updates);
   my $st = $self->{db}->prepare(qq{
      update $self->{table} set $updates $self->{byKey}
   });
   $st->execute(@values, $self->order_keys($key));
   $st->finish;
}

1;
