#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# afb 2/96, etter 2/97, afb 12/1999
#------------------------------------------------------------------------------
# Locking operations based on fcntl
#------------------------------------------------------------------------------

package TBD::Locker;
use TBI;
use Carp;
use English;
use Fcntl;
use strict;
use warnings;

our $VERSION = "0.01";
our @ISA = qw(TBI);

=head1 NAME

TBD::Locker - locking operations

=head1 SYNOPSIS

   use TBD::Locker;
   our @ISA = qw(... TBD::Locker ...);

=head1 DESCRIPTION

TBD::Locker provides fcntl-based lock operations that may be used by
derived TBI implementations that rely on ordinary files.

=over 4

=cut

=item $table->lock($fh)

locks the file associated with $fh for write accesses.  If someone else
holds a lock on this file, this operation blocks until the previously
existing lock is released.

=cut

sub lock {
   my ($table, $fh) = @_;
   my $fcmd    = &F_SETLKW;
   my $fcntl_t = "ssllllllll";
   my @array   = ();
   my $fcntl;

   $array[0] = &F_WRLCK;
   $array[1] = 0;
   $array[2] = 0;
   $array[3] = 0;

   ($fcntl = pack($fcntl_t, @array)) ||
      croak "locking " . $table->name . "failed: pack: $!";
   fcntl($fh, $fcmd, $fcntl) ||
      croak "locking " . $table->name . "failed: fcntl: $!";
}

=item $table->unlock($fh)

unlocks the file associated with $fh.

=cut

sub unlock {
   my ($table, $fh) = @_;
   my $fcmd    = &F_SETLK;
   my $fcntl_t = "ssllllllll";
   my @array   = ();
   my $fcntl;

   $array[0] = &F_UNLCK;
   $array[1] = 0;
   $array[2] = 0;
   $array[3] = 0;

   ($fcntl = pack($fcntl_t, @array)) ||
      croak "unlocking " . $table->name . "failed: pack: $!";
   fcntl($fh, $fcmd, $fcntl) ||
      croak "unlocking " . $table->name . "failed: fcntl: $!";
}

=item $table->islocked($fh)

returns true if the file associated with $fh is currently locked.

=cut

sub islocked {
   my ($table, $fh) = @_;
   my $fcmd    = &F_GETLK;
   my $fcntl_t = "ssllllllll";
   my @array   = ();
   my $fcntl;

   return 0 if $fh == undef;

   $array[0] = &F_WRLCK;
   $array[1] = 0;
   $array[2] = 0;
   $array[3] = 0;

   ($fcntl = pack($fcntl_t, @array)) ||
      croak "checking lock for " . $table->name . "failed: pack: $!";
   fcntl($fh, $fcmd, $fcntl) ||
      croak "checking lock for " . $table->name . "failed: fcntl: $!";
   (@array = unpack($fcntl_t, $fcntl)) ||
      croak "checking lock for " . $table->name . "failed: unpack: $!";

   $array[0] != &F_UNLCK;
}

=back

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/borchert/">Andreas Borchert</A>
with contributions of
<A HREF="http://www.mathematik.uni-ulm.de/sai/etter/">Mathias Etter</A>.

=for text
Andreas Borchert with contributions of Mathias Etter.

=for man
Andreas Borchert with contributions of Mathias Etter.

=cut

1;
