#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# afb 2/96
#------------------------------------------------------------------------------
# General scanning algorithm for tables
#------------------------------------------------------------------------------

package TBD::Scanner;
use Carp;
use English;
use TBI;
use strict;
use warnings;

our $VERSION = "0.02";
our @ISA = qw(TBI);

=head1 NAME

TBD::Scanner - a general scanning algorithm for tables

=head1 SYNOPSIS

   use TBD::Scanner;
   @ISA = qw(... TBD::Scanner ...);

=head1 DESCRIPTION

TBD::Scanner provides a general scanning algorithm for tables that are
represented by text files that relies on following internal variables:

=over 4

=item $table->{ifs}

Input field separator that is interpreted as regular expression.

=item $table->{irs}

Input record separator which is to be given as string.

=item $table->{comment}

A string (or regular expression without leading ^)
which matches comment lines which are to be ignored in
the file. This string may be empty if no comments are used.

=back

Following operation is defined:

=over 4

=item $table->scan($handle)

loads the table from the given input handle into memory by filling
$table->{table}.  Input is parsed under consideration of $table->{ifs}
(input field seperator that will be interpreted as regular expression)
and $table->{irs} (input record separator).  Empty lines and lines
beginning with $table->{comment} are ignored.

The fields are expected in the same order as returned by $table->fields.
The table $table->{table} is (in conformance to TBD::CachingTable)
represented as hash table consisting of references to hash tables
whose keys are the field names. Keys consisting of multiple fields
are joined together using $table->{ofs} to provide a string for
hash table keys.

All leading comments and empty lines are stored into $table->{header}
(as list of records without record separators).  Comments and empty
lines which appear later are lost.

True is returned in case of success, false in case of failures.

=back

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/borchert/">Andreas Borchert</A>

=for text
Andreas Borchert

=for man
Andreas Borchert

=cut

sub scan {
   my ($table, $handle) = @_;
   my $comment = $table->{comment};
   my $ifs = $table->{ifs};
   my $start = 1;
   my @keyfields = $table->keyfields;
   my @fieldnames = $table->fields;

   my $oh = select($handle); $RS = $table->{irs}; select($oh);
   $table->{table} = {};
   $table->{header} = [];
   while(<$handle>) {
      chomp;
      if (defined($comment) && $comment && /^$comment/) { # comment line
	 push(@{$table->{header}}, $_) if $start;
	 next;
      }
      my @fields = split /$ifs/;
      if (@fields == 0) {		# empty line
	 push(@{$table->{header}}, $_) if $start;
	 next;
      }
      $start = 0;
      my %fields = ();
      foreach my $field (@fieldnames) {
	 last if @fields == 0;
	 $fields{$field} = shift @fields;
      }
      my $key = join($table->{ofs}, @fields{@keyfields});
      $table->{table}->{$key} = \%fields;
   }
   return 0 if $handle->error;
   return 1;
}

1;
