#------------------------------------------------------------------------------
# University of Ulm, Faculty of Mathematics and Economics
# afb 6/98
#------------------------------------------------------------------------------
# general abstraction for database tables
#------------------------------------------------------------------------------
# $Id$
#------------------------------------------------------------------------------
# $Log$
#------------------------------------------------------------------------------

package TBI;

use Exporter;
use strict;
use warnings;

our $VERSION = "0.09";
our @ISA = qw(Exporter);

#------------------------------------------------------------------------------

=head1 NAME

TBI - abstract table interface

=head1 SYNOPSIS

   use TBI;

   $t = TBI->open($name);
   @tables = TBI->tables();
   $rc = TBI->exists($name);
   TBI->register($name, $package, \%params);

   $name = $t->name;
   @names = $t->fields;
   @names = $t->keyfields;

   $rc = $t->exists($key);
   %fields = $t->fetch($key);
   @fields = $t->get($key, @fieldnames);
   $field = $t->getfield($key, $fieldname);
   @keys = $t->keys;
   @keys = $t->select(%equations);

   $t->add($key, %entry);
   $t->delete($key);
   $t->modify($key, %changes);

   $t->close;

=head1 DESCRIPTION

I<TBI> provides a convenient abstract interface for various concrete table
implementations including all I<DBI> drivers.  Each table is defined by
an identifying name, a set of named fields, and a subset thereof that
constitutes the primary key. Field values are scalar string values.

=head2 Class Methods

I<TBI> maintains a global name space for tables. Any implementation
or initialization procedure is free to register more tables but it
should be taken care of possible name clashes -- usually the first
entry should remain undisturbed.

=over 4

=item open

if the given table name has been successfully opened before,
its pointer is returned. Otherwise, the table is newly opened.
Upon success, a valid pointer is returned, otherwise undef.

=item tables

returns all table names known so far.

=item exists

checks whether a table with the given name is present.

=item register

registers a table for the given name. Multiple registrations for
the same name are permitted, causing the older registration to be lost.
Usually this method is called by implementors or initialization procedures
that set up your data dictionary (see below).

=back

=head2 Methods of Table Objects

Following operations are supported:

=over 4

=item name

return the identifying name of the table as string.

=item fields

return the names of all fields as a list of strings.

=item keyfields

return the names of all primary key fields as a list of strings.

=item fetch($key)

return the referenced table entry as
hash table where the field names serve as keys.
For example,

   %fields = $members->fetch("borchert");

is expected to return all fields of the selected table entry.
In case of multiple key fields, a pointer to an associative array has to
be given:

   %fields = $documents->fetch({group => '1', shortname => '13'});

=item exists($key)

return true if an entry with the given key exists in the table.

=item get($key, @fieldnames)

return the specified fields of the selected table entry as list of
scalars.

=item getfield($key, $fieldname)

return the specified field of the selected table entry.

=item keys

return all keys of the table as list.
This allows to iterate through all table entries:

   foreach $key ($members->keys) {
      # operate on $members->fetch($key)
   }

=item select(%equations)

return a list of keys of table entries for which the given equations hold.
Following example retrieves all documents of an author who belongs
to a given group:

   @keys = $authors->select(login => $login, groupname => $groupname);

=item add($key, %entry)

add an entry to the table with the given key.
Note that the key must not have been used previously.

=item delete($key)

delete the entry with the given key.

=item modify($key, %entry)

modify the given components of the already existing key.
Note that %entry does not need to give all component-value pairs.
For example,

   $accounts->modify("borchert", name => 'Andreas Franz Borchert');

modifies just the name of borchert but not the other components.

=item close

closes the given table. Note that this method is seldomly called directly
because I<TBI> closes tables implicitly if they are subject to the
garbage collector. Special care has to be taken in case of multiple
openings of the same table name because they all share the same table
object.  The first close would make the table inaccessible by the other
parties using that table object.

=back

=head2 Guide for implementors

Implementations of the I<TBI> abstraction (sometimes also called
drivers) are conventionally put under the I<TBD> hierarchy. This
is similar to the structure of the I<DBI> interface where drivers
are put under the I<DBD> hierarchy.

I<TBI> provides the constructor I<new> that is called by the I<open>
method with the parameters supplied previously by I<register>.
I<new> starts a two-phase initialization procedure using the
methods I<initialize1> and I<initialize2>. Together with I<close>,
all these methods should be executed through-out the whole @ISA-chain
using following patterns:

   sub initialize1 {
      my ($self, %attributes) = @_;

      $self->SUPER::initialize1(%attributes);

      # early initialization code
   }

   sub initialize2 {
      my ($self) = @_;

      # late initialization code

      $self->SUPER::initialize2();
   }

   sub close {
      my ($self) = @_;

      # clean up code

      $self->SUPER::close();
   }

Due to the two-phase initialization it is possible to embed the
initialization code of the more derived classes into the initialization
of the more general packages. Note that I<close> may be invoked
explicitly or implicitly by the I<DESTROY> method of I<TBI>.

Some of the methods have default implementations in I<TBI>:
I<name>, I<fields>, I<keyfields>, and I<get>.

Table objects are represented by hashes whose keys are used
like field names. Following field names of table objects are
predefined or used by I<TBI>:

=over 4

=item closed

is set to 1 by the I<close> method of I<TBI> to avoid repeated executions
of cleanup code. Otherwise cleanup might occur twice, i.e. once on an
explicit I<close> and once due to I<DESTROY>.

=item tablename

is expected be set by the implementation to hold the name of the table.

=item fields

must be set by the implementation as a pointer to an ordered list
of field names.

=item keyfields

is expected to be a reference to an ordered list of field names
belonging to the primary key.

=back

=head2 Registration

During its initialization, I<TBI> tries to load a module named
I<TBI_Scanner> and invokes the class method I<scan> of it. This
hook allows to register a base set of tables of your application.
If I<TBI_Scanner> is not found, I<TBI> starts with an empty table set.

Implementations are free to add more tables (e.g. all other tables of a
database if the first one gets opened) if and only if the global variable
$TBI::AUTO_REG is true. Applications should set this variable to 0 if
they want to keep full control over the name space of I<TBI>.

=head1 DIAGNOSTICS

Some of the operations may croak in case of fatal errors, for example
on invalid field names, or on trying to add tuples with an already
existing primary key.

=head1 AUTHOR

=for html
<A HREF="http://www.mathematik.uni-ulm.de/sai/borchert/">Andreas Borchert</A>

=for text
Andreas Borchert

=for man
Andreas Borchert

=cut

#------------------------------------------------------------------------------

our $AUTO_REG = 1;

my %tables = ();

eval q{
   package TBI::_firesafe;
   use TBI_Scanner;

   TBI_Scanner->scan();
};

#------------------------------------------------------------------------------

sub open {
   my ($package, $name) = @_;
   my $table = undef;
   if (defined($tables{$name})) {
      unless (defined($table = $tables{$name}->{handle})) {
	 $table = $tables{$name}->{package}->
		  new(%{$tables{$name}->{params}});
      }
   }
   return $table;
}

sub register {
   my ($self, $name, $package, $params) = @_;
   unless (defined($tables{$name})) {
      $tables{$name} = {};
   }
   $tables{$name}->{package} = $package;
   $tables{$name}->{params} = $params;
}

sub tables {
   return sort keys %tables;
}

sub exists {
   my $table = shift;
   return exists $tables{$table};
}

sub new {
   my $package = shift;
   my $self = bless {closed => 0}, $package;
   $self->initialize1(@_);
   $self->initialize2();
   my $name = $self->name();
   if (defined($tables{$name})) {
      $tables{$name}->{handle} = $self;
   } else {
      $tables{$name} = {handle => $self};
   }
   return $self;
}

#------------------------------------------------------------------------------

sub initialize1 { }

sub initialize2 { }

sub close {
   my $self = shift;
   $self->{closed} = 1;
   delete $tables{$self->name()}->{handle};
}

sub DESTROY {
   my $self = shift;
   $self->close() unless $self->{closed};
}

sub name {
   my $self = shift;
   return $self->{tablename};
}

sub fields {
   my $self = shift;

   return @{$self->{fields}};
}

sub keyfields {
   my $self = shift;

   return @{$self->{keyfields}};
}

sub get {
   my ($self, $key, @fieldnames) = @_;
   my (@fields, $field);

   my %fields = $self->fetch($key);
   foreach $field (@fieldnames) {
      push(@fields, $fields{$field});
   }
   return @fields;
}

sub getfield {
   my ($self, $key, $fieldname) = @_;
   my ($field) = $self->get($key, $fieldname);
   return $field;
}

1;
